# Author: Ivan Canay | Northwestern University | January 2015
#---------------------------------------------------------------------------------------
# X-plain: Functions for Inference under covariate-adaptive Randomization
# Latest Revision: April 7, 2017: I changed the way SFECAP computes standard errors.
#                  April 9, 2017: I added the new t statistic with new SE
#                 April 10, 2017: made changes to newt.stat to handle NAs better 
#                 April 25, 2017: CAP SFEP are both for adjusted test statistics too now
#                 April 27, 2017: Changed the function SFE.var and Corrected mistake in CAP SFE and CAP SFE adj
#---------------------------------------------------------------------------------------
require("compiler")
require("MASS")
library(exactRankTests)
library(coin)
require("sandwich")
require("lmtest")
require("VGAM")
require("matrixStats")

# We use the R compilter to improve the speed of some functions.
#-------------------------------------------------------------------
# List of Functions:
# 1: gen.baseline.one: generates potential outcomes and 1 covariate
# 2: form.strata.one:  forms strata in the case of 1 covariate
# 3: dgp.srs: Random assignments using SRS
# 4: dgp.wei: Random assignments using Wei's design
# 5: dgp.bcd: Random assignments using Bias Coin Design
# 6: dgp.sbr: Random assignments using Stratified Block Randomization 
# 7: t.stat: computes the two-sample t-statistic
# 8: P.test: Computes a standard permutation test
# 9: A.CAP: Computes the set of Covariate adaptive permutations
# 10: CAP.test: Computes the covariate adaptive permutation (CAP) test
# 11: sq.H : computes sigma squared H in different regression ways
# 12: SFE.var: computes the new estimators for the variance of SFE
# 13: CAP.SFE: computes the CAP test on the SFE test statistic
# 14: shao.boot: Computes the bootstrap approach in Shao et al (2010)
# 15: true.values: computes population values for the key parameters of each model 
# 16: adjt.stat: computes t statistic with new standard errors
#------------------------------------------------------------------

#-------------------------------------------------------------------
gen.baseline.one <- function(n,model,theta,gamma,sg1){
#-------------------------------------------------------------------
# X-plain: Generates covariates (W) and potential outcomes (Y.1,Y.0)
# using different functions m() and epsilons.
#-------------------------------------------------------------------
  mu.1<- theta; # We normalized mu.0=0 so theta=mu.1

  # NOTE: Model 1 has always been Model 1. 
  if (model==1){ #linear model with normal epsi and beta W
    epsi.1<- sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- rnorm(n);
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    # function m() of each potential outcomes (check E[m(W)]=0)
    #m0    <- gamma*W; m0<-m0 - mean(m0);
    m0    <- gamma*W;
    m1    <- m0;
    Y.1   <- mu.1 + m1 + epsi.1;
    Y.0   <- m0 + epsi.0;
  }
  # Note: Model 2 used to be model 7 (which in turn is Model 2 in Expensive CAR)
  if (model==2){ #linear/non-linear but heterogeneous | Model 2 in Expensive CAR
    # This model with sg1=2 and pi=1/3 should introduce distortion for HOMO-SFE
    # If gamma=4 sigmaH becomes small. gamma=1 or 2 are fine. 
    epsi.1<- sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- rnorm(n);
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    # function m() of each potential outcomes (check E[m(W)]=0)
    m1    <- gamma*W;
    m0    <- gamma*(-((log(W+3))*(W<1/2)-0.5602138));
    Y.1   <- mu.1 + m1 + epsi.1;
    Y.0   <- m0 + epsi.0;
  }
  # NOTE: model 3 below used to be model 5 in the first submission (and was model 4 in previous rounds of the revision)
  if (model==3){ #Non-linear, t-dist, homogeneous 
    W     <- runif(n,-2,2);
    epsi.1<- W^2*sqrt(sg1)*rt(n,3)/3; # sg1: sigma squared epsi 1
    epsi.0<- W^2*rt(n,3)/3; # divide by 3 to make var = 1.
    # function m() of each potential outcomes (check E[m(W)]=0)
    m0    <-gamma*((2-W^2)*(W+1<0)+W^2*(W+1>=0)*(W<1)+(2-W^2)*(W>=1));
    m1    <- m0;
    Y.0   <- m0 + epsi.0;
    Y.1   <- mu.1 + m1 + epsi.1;
  }
  # Note: Model 4 below is model 4 in expensive CAR | Used to be model 5 in the previous rounds of simulations of the revision
  if (model==4){ #Non-linear, heterogeneous, t-distribution, homogeneous.| Model 4 in Expensive CAR | new model 4 in Inference CAR
    W     <- runif(n,-2,2);
    epsi.1<- abs(W)*sqrt(sg1)*rt(n,3)/3; # sg1: sigma squared epsi 1
    epsi.0<- abs(W)*rt(n,3)/3; # divide by 3 to make var = 1.
    # function m() of each potential outcomes (check E[m(W)]=0)
    m0    <- gamma*(W*(W+1<0)+W^2*(W+1>=0)*(W<=1)+W*(W>1)-0.1724343);
    m1    <- gamma*(W^2*(W+1<0)+W*(W+1>=0)*(W<=1)+W^2*(W>1)-1.158188);
    Y.0   <- m0 + epsi.0;
    Y.1   <- mu.1 + m1 + epsi.1;
  }
  # ---------------------------------------------------------------------------------------------------------------------- 
  # The models below are taken from the Expensive CAR paper or are old models we used in older versions of the paper
  # ----------------------------------------------------------------------------------------------------------------------
   if (model==6){ #Heterogeneous/Hteroskedastic model (This was used in the first version of paper)
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    epsi.1<- W^2*sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- W^2*rnorm(n);
    # function m() of each potential outcomes (check E[m(W)]=0)
    #m0    <- sin(gamma*W); m0<- m0 - mean(m0);
    #m1    <- sin(gamma*W)+sqrt(W+2.25)-1.4556; m1<- m1 - mean(m1);
    m0    <- sin(gamma*W);
    m1    <- sin(gamma*W)+sqrt(W+2.25)-1.4556;
    Y.0   <- m0 + epsi.0;
    Y.1   <- mu.1 + m1 + epsi.1;
  }
  # Note: Model 7 below used to be MODEL 2 in our first submission. 
  if (model==7){ #sin model with normal epsi and beta W
    epsi.1<- sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- rnorm(n);
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    # function m() of each potential outcomes (check E[m(W)]=0)
    #m0    <- sin(gamma*W); m0<- m0 - mean(m0);
    m0    <- sin(gamma*W);
    m1    <- m0; 
    Y.1   <- mu.1 + m1 + epsi.1;
    Y.0   <- m0 + epsi.0;
  }
  #NOTE: Model 8 used to be MODEL 3 in our first submission. 
  if (model==8){ #Heterogeneous model with sin and box-cox
    epsi.1<- sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- rnorm(n);
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    # function m() of each potential outcomes (check E[m(W)]=0)
    #m0    <- sin(gamma*W); m0<- m0 - mean(m0);
    #m1    <- sin(gamma*W)+sqrt(W+2.25); m1<- m1 - mean(m1);
    m0    <- sin(gamma*W);
    m1    <- sin(gamma*W)+sqrt(W+2.25)-1.4556;
    Y.0   <- m0 + epsi.0;
    Y.1   <- mu.1 + m1 + epsi.1;
  }
   if (model==9){ #Like model 2 but heteroskedastic | Model 3 in Expensive CAR
    W     <- sqrt(20)*(rbeta(n,2,2)-1/2); # recentered 0 and re-scaled 1
    epsi.1<- sqrt(abs(W))*sqrt(sg1)*rnorm(n); # sg1: sigma squared epsi 1
    epsi.0<- sqrt(abs(W))*rnorm(n);
     # function m() of each potential outcomes (check E[m(W)]=0)
    m1    <- gamma*W;
    m0    <- gamma*(-((log(W+3))*(W<1/2)-0.5602138));
    Y.1   <- mu.1 + m1 + epsi.1;
    Y.0   <- m0 + epsi.0;
  }
   
  # List of returns
  list(Y.0=Y.0,Y.1=Y.1,W=W,e0=epsi.0,e1=epsi.1,m.1=m1,m.0=m0,n=n,model=model)
}
#-------------------------------------------------------------------
form.strata.one <- function(baseline,num.strata){
#-------------------------------------------------------------------
# X-plain: Generates strata indicators from covariates (W).
#         - W: covariates (must be scalar)
#         - num.strata: the number of strata to be formed
#         - model: model in baseline.
#-------------------------------------------------------------------
model<-baseline$model;
n <- baseline$n;

if (model<=2){# W is beta in [-sqrt(20)/2,sqrt(20)/2]
  W<-baseline$W;
  bounds<-seq(-2.25,2.25,length.out = num.strata+1); # careful with bounds
  I.S   <- matrix(0,n,num.strata);
  for (s in 1:num.strata){
    I.S[,s]<- (W>bounds[s])*(W<=bounds[s+1]);
  }     
}
if (model==3|model==4){# W is uniform [-2,2]
  W<-baseline$W;
  bounds<-seq(-2,2,length.out = num.strata+1); # careful with bounds
  I.S   <- matrix(0,n,num.strata);
  for (s in 1:num.strata){
    I.S[,s]<- (W>bounds[s])*(W<=bounds[s+1]);
  }     
}
if (model>=5){
  W<-baseline$W;
  bounds<-seq(-2.25,2.25,length.out = num.strata+1); # careful with bounds
  I.S   <- matrix(0,n,num.strata);
  for (s in 1:num.strata){
    I.S[,s]<- (W>bounds[s])*(W<=bounds[s+1]);
  }  
}

return(I.S)
}
#-------------------------------------------------------------------
dgp.srs <- function(baseline,I.S,pi=1/2){
#-------------------------------------------------------------------
# X-plain: Generates treatment assignments and outcomes using SRS.
#         - baseline: structure from `gen.baseline' 
#         - I.S: matrix from `form.strata' 
#         - pi: target proportion of n1/n - could be a vector
#-------------------------------------------------------------------
  n   <-baseline$n;
  num.strata <- ncol(I.S);

  # Compute the vector of pis
  dim.pi<-length(pi);
  if (dim.pi==1){pi<-rep(pi,n)}
  if (dim.pi>1){
    new.pi<-rep(0,n);
    gap<-floor(num.strata/dim.pi);
    for (j in 1:dim.pi){
      lb<-(j-1)*gap+1;
      ub<- j*gap;
      new.pi<-new.pi+rowSums(I.S[,lb:ub])*pi[j];
    }
    new.pi[new.pi==0]<-pi[dim.pi];
    pi<-new.pi;
  }

  A   <-(runif(n)<=pi); # treatment Assignment
  A   <-as.numeric(A);
  Y   <- baseline$Y.1*A + baseline$Y.0*(1-A);
  A.S <- matrix(c(A)*I.S,n,num.strata);

  data<-data.frame(Y=Y,W=baseline$W,pi=pi,A=A,I.S=I.S,AS=A.S);
  return(data);
}
#-------------------------------------------------------------------
dgp.wei <- function(baseline,I.S,pi=1/2){
#-------------------------------------------------------------------
# X-plain: Generates treatment assignments and outcomes using Wei's design.
#         - baseline: structure from `gen.baseline' 
#         - I.S: matrix from `form.strata' 
#         - pi: target proportion of n1/n - could be a vector.
#-------------------------------------------------------------------
  num.strata <- ncol(I.S);
  n   <-baseline$n;
  A   <- cbind(rep(0,n));    # treatment Assignment
  D.s <- rep(0,num.strata);  # inbalance in each stratum s
  n.s <- rep(0,num.strata);  # sample size in each s

  # Compute the vector of pis
  dim.pi<-length(pi);
  if (dim.pi==1){pi<-rep(pi,n)}
  if (dim.pi>1){
    new.pi<-rep(0,n);
    gap<-floor(num.strata/dim.pi);
    for (j in 1:dim.pi){
      lb<-(j-1)*gap+1;
      ub<- j*gap;
      new.pi<-new.pi+rowSums(I.S[,lb:ub])*pi[j];
    }
    new.pi[new.pi==0]<-pi[dim.pi];
    pi<-new.pi;
  }
  
  for (j in 1:n){
    index   <- which(I.S[j,]==1);
    n.s[index]<-n.s[index]+1;
    phi     <-(1-D.s[index]/n.s[index])*pi[j];
    A[j]    <-(runif(1)<=phi);
    D.s[index]<-(1/pi[j])*sum(A*(I.S[,index]==1))-n.s[index]; 
  }
  A<-as.numeric(A);
  Y   <- baseline$Y.1*A + baseline$Y.0*(1-A);

  A.S <- matrix(c(A)*I.S,n,num.strata);

  data<-data.frame(Y=Y,W=baseline$W,pi=pi,A=A,I.S=I.S,AS=A.S);
  return(data);
}
#-------------------------------------------------------------------
dgp.bcd <- function(baseline,I.S,pi=1/2,factor=1/2){
#-------------------------------------------------------------------
# X-plain: Treatment assignments and outcomes using Bias Coin design.
#         - baseline: structure from `gen.baseline' 
#         - I.S: matrix from `form.strata' 
#         - pi: target proportion of n1/n - could be a vector
#-------------------------------------------------------------------
  num.strata <- ncol(I.S);
  n   <-baseline$n;
  A   <- cbind(rep(0,n));    # treatment Assignment
  phi <- A;
  D.s <- rep(0,num.strata);  # inbalance in each stratum s
  n.s <- rep(0,num.strata);  # sample size in each s

  # Compute the vector of pis
  dim.pi<-length(pi);
  if (dim.pi==1){
    if (pi>=0.5){lambda<- pi + (1-pi)*factor};
    if (pi<0.5){lambda <- 1-pi + pi*factor};
    pi <-rep(pi,n);
    lambda<-rep(lambda,n);
  }
  if (dim.pi>1){
    new.pi<-rep(0,n);
    gap<-floor(num.strata/dim.pi);
    for (j in 1:dim.pi){
      lb<-(j-1)*gap+1;
      ub<- j*gap;
      new.pi<-new.pi+rowSums(I.S[,lb:ub])*pi[j];
    }
    new.pi[new.pi==0]<-pi[dim.pi];
    pi<-new.pi;
    lam1<-pi+(1-pi)*factor;
    lam2<-1-pi+pi*factor;
    lambda<-(pi>=0.5)*lam1+(pi<0.5)*lam2;
  }
 
  for (j in 1:n){
    index   <- which(I.S[j,]==1);
    n.s[index]<-n.s[index]+1;
    phi  <- lambda[j]*(D.s[index]<0)+pi[j]*(D.s[index]==0)+(1-lambda[j])*(D.s[index]>0);
    A[j]    <-(runif(1)<=phi);
    D.s[index]<-(1/pi[j])*sum(A*(I.S[,index]==1))-n.s[index]; 
  }
  A<-as.numeric(A);
  Y   <- baseline$Y.1*A + baseline$Y.0*(1-A);

  A.S <- matrix(c(A)*I.S,n,num.strata);

  data<-data.frame(Y=Y,W=baseline$W,pi=pi,A=A,I.S=I.S,AS=A.S);
  return(data);

}
#-------------------------------------------------------------------
dgp.sbr <- function(baseline,I.S,pi=1/2){
#-------------------------------------------------------------------
# X-plain: Treatment assignments and outcomes using PBWS.
#         - baseline: structure from `gen.baseline' 
#         - I.S: matrix from `form.strata' 
#         - pi: target proportion of n1/n
#-------------------------------------------------------------------
  num.strata <- ncol(I.S);
  n   <-baseline$n;
  A   <- cbind(rep(0,n));    # treatment Assignment

  # Compute the vector of pis
  dim.pi<-length(pi);
  if (dim.pi==1){pi<-rep(pi,n)}
  if (dim.pi>1){
    new.pi<-rep(0,n);
    gap<-floor(num.strata/dim.pi);
    for (j in 1:dim.pi){
      lb<-(j-1)*gap+1;
      ub<- j*gap;
      new.pi<-new.pi+rowSums(I.S[,lb:ub])*pi[j];
    }
    new.pi[new.pi==0]<-pi[dim.pi];
    pi<-new.pi;
  }
   
  for (k in 1:num.strata){
    index   <- which(I.S[,k]==1);
    pi.s    <- mean(pi[index],na.rm=TRUE);
    pi.s[is.na(pi.s)]<-0;
    ns      <- length(index);
    # pick a random permutation of 1s and 0s. 
    A[index]<- sample(c(rep(1,floor(pi.s*ns)),rep(0,ns-floor(pi.s*ns))));
  }
  A<-as.numeric(A);
  Y   <- baseline$Y.1*A + baseline$Y.0*(1-A);

  A.S <- matrix(c(A)*I.S,n,num.strata);

  data<-data.frame(Y=Y,W=baseline$W,pi=pi,A=A,I.S=I.S,AS=A.S);
  return(data);
}
#-------------------------------------------------------------------
t.stat <- function(data,theta=0){
#-------------------------------------------------------------------
# X-plain: compute the usual t-statistic of diff in means
#-------------------------------------------------------------------
Y   <-data$Y;
A   <-data$A;
n 	<- length(A);
n.1	<- sum(A); 
n.0	<- n-n.1;

Y1.bar   <-sum(Y*A)/n.1; 
Y0.bar   <-sum(Y*(1-A))/n.0;
sigma2_1 <-sum((Y-Y1.bar)^2*A)/n.1;
sigma2_0 <-sum((Y-Y0.bar)^2*(1-A))/n.0;

t.stat<-abs(Y1.bar-Y0.bar-theta)/sqrt(sigma2_1/n.1+sigma2_0/n.0);
}

#-------------------------------------------------------------------
pre.P.test <- function(data,theta=0,M=500,alpha=0.05){
#-------------------------------------------------------------------
# X-plain: compute the Naive permutation test for the two-sample t-statistic
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - theta: value of theta under the null
#         - M: number of elements from the set of all permutations.  
#         - alpha: significance level of the test  
#-------------------------------------------------------------------
# First compute Observed test statistic
  A   <-data$A; Y   <-data$Y;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0	<- n-n.1;
  Y1.bar   <-sum(Y*A)/n.1; 
  Y0.bar   <-sum(Y*(1-A))/n.0;
  sigma2_1 <-sum((Y-Y1.bar)^2*A)/n.1;
  sigma2_0 <-sum((Y-Y0.bar)^2*(1-A))/n.0; 
  
  ObsT<-abs(Y1.bar-Y0.bar-theta)/sqrt(sigma2_1/n.1+sigma2_0/n.0);
  
  # Now Compute permuted test statistic. First get all transformations
  ind   <- (runif(n*M)>1/2);
  ind   <- matrix(ind,n,M);
  A.per <- abs(A-ind); # every column is 1 permutation
  Y1.per <- Y*A.per; # every column is 1 permutation
  Y0.per <- Y*(1-A.per); # every column is 1 permutation
  n.1    <- colSums(A.per); n.0<-n-n.1; # number of ones and zeros per column
  bar_1  <- colSums(Y1.per)/n.1; # average per set of permutations
  bar_0  <- colSums(Y0.per)/n.0;
  vec.bar_1 <-t(matrix(rep(bar_1,n),M,n)); # useful to compute variance
  vec.bar_0 <-t(matrix(rep(bar_0,n),M,n));
  
  sig.sq_1   <- colSums((Y1.per-vec.bar_1)^2*A.per)/n.1;
  sig.sq_0   <- colSums((Y0.per-vec.bar_0)^2*(1-A.per))/n.0;
    
  # Compute the new test statistic over permuted data
  nume    <- abs(bar_1-bar_0-theta);
  deno    <- sqrt(sig.sq_1/n.1 + sig.sq_0/n.0);
  NewT    <- sort(nume/deno);    # sort the vector of Test Stat  
  k       <- M-floor(M*alpha);   # find index for quantile
  Mplus   <- sum(NewT>NewT[k]);  # M+ from LR book
  M0      <- sum(NewT==NewT[k]); # M0 from LR book
  prob    <- (M*alpha-Mplus)/M0; # term a(x) from LR book
  # randomized test 
  rule <- (ObsT>NewT[k]) + (ObsT==NewT[k])*(runif(1)<=prob);  
  # non-randomized test
  Nrule <- (ObsT>NewT[k]);
  ties <- sum(ObsT==NewT[k]);
  
  # List of returns
  list(stat=ObsT,rule=rule, Nrule=Nrule, cv=NewT[k],ties=ties)  
}
#-------------------------------------------------------------------
P.test <- cmpfun(pre.P.test);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
A.CAP <- function(data,M=500){
#-------------------------------------------------------------------
# X-plain: Set of Covariate Adaptive Permutations. 
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - M: number of elements from the set of all permutations.  
#-------------------------------------------------------------------
  I.S   <-data[,grepl("I.S", names(data))]
  num.strata<- ncol(I.S);
  Y   <-data$Y;
  A   <-data$A;
  n   <- length(A);
   
  # Now Compute the set of CAP permutations
  A.per      <- matrix(0,n,M); 
   
  for (k in 1:num.strata){
    index   <- which(I.S[,k]==1);
    A.s     <-A[index];
    # Here use "replicate" to get a n x M matrix of Permutations.
    if (length(index)>0){
        A.per[index,]<- replicate(M,sample(A.s));
    }
  }
  # include the identity element
  A.per[,1]<-A;

  A.CAP<-A.per;
}

#-------------------------------------------------------------------
sq.H <- function(SFE,SAT,I.S,pi,method=1,A=0){
#-------------------------------------------------------------------
# X-plain: Computes \sigma^2_H 
#         - SFE: SFE regression  
#         - SAT: SAT regression
#         - I.S: Strata dummies.
#         - pi: n dimensional vector of pi(s)
#         - method: 1(Sat only), 2(residuals no A), 3(residuals with A) 
#-------------------------------------------------------------------
  num.strata<- ncol(I.S);
  n   <- length(pi);
  n.s <- colSums(I.S);
  p.s <- colMeans(I.S);
  pi.s<- colMaxs(I.S*pi);
  u.sfe   <- residuals(SFE);
  u.sat   <- residuals(SAT);
  b.sfe   <- coef(summary(SFE))[2,1];
  b.sat   <- coef(summary(SAT))[,1]; # This is a vector

  # Compute theta hat - SAT
    weights<-c(rep(0,num.strata),colMeans(I.S));
    R      <- matrix(weights[!is.na(coef(SAT))]);
    theta  <- sum(R*b.sat);

if (method==1){
  sq.H <- sum(R*(b.sat-theta)^2);
  }

if (method==2){
  # First and second Terms
  term1.s <- sqrt(pmax(colSums((u.sfe^2-u.sat^2)*I.S)/pmax(n.s*pi.s*(1-pi.s),0.000001),0));
  term2   <- (b.sfe-theta)*sum(p.s*pi.s*(1-pi.s));
  # Sigma^2 H
  sq.H <- sum(p.s*(term1.s+term2)^2);
  }
if (method==3){
  # First and second Terms
  term1.s <- sqrt(pmax(colSums((u.sfe^2-u.sat^2)*I.S*A)/pmax(n.s*pi.s*(1-pi.s)^2,0.000001),0));
  term2   <- (b.sfe-theta)*sum(p.s*pi.s*(1-pi.s));
  # Sigma^2 H
  sq.H <- sum(p.s*(term1.s+term2)^2);
  }

  return(sq.H)
   
}

#-------------------------------------------------------------------
pre.shao.boot <- function(data,theta=0,alpha=0.5,p=2/3,B=500){
#-------------------------------------------------------------------
# X-plain: computes the bootstrap t-stat in Shao et al (2010)
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - theta: value of theta under the null
#         - p: parameter for the BCD
#         - B: number of bootstrap replications. 
#-------------------------------------------------------------------
  num.strata<- ncol(data[,grepl("I.S", names(data))]); 
  A   <-data$A; 
  Y   <-data$Y;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y1.bar   <-sum(Y*A)/n.1; 
  Y0.bar   <-sum(Y*(1-A))/n.0;
  
  boo.nume  <- matrix(0,B,1);
  phi        <- c(p,1/2,1-p);

  for (b in 1:B){
    boo.data   <- data[sample(n,size=n,replace=TRUE),];
    boo.I.S     <-boo.data[,grepl("I.S", names(boo.data))];
    n.s <- rep(0,num.strata);  # sample size in each s
    D.s <- rep(0,num.strata);  # inbalance in each stratum s
    boo.A      <- matrix(NA,n,1); # Bootstrapped treatment assignment
  
    # Let's try loop over strata
    for (k in 1:num.strata){
      index   <- which(boo.I.S[,k]==1);
      ns     <- length(index);
      D.s     <- 0;
      if (ns>0){
        for (i in 1:ns){
          boo.A[index[i]] <-(runif(1)<=phi[sign(D.s)+2]);
          D.s             <-D.s + 2*boo.A[index[i]]-1;    
        }
      }
    }
  
    # loop over n very slow (repeats "which" too much - AVOID) 
    # for (j in 1:n){
    #   index   <- which(boo.I.S[j,]==1);
    #   n.s[index]<-n.s[index]+1;
    #   boo.A[j]    <-(runif(1)<=phi[D.s[index]+2]);
    #   D.s[index]<-2*sum(boo.A*(boo.I.S[,index]==1))-n.s[index]; 
    # }

    # Store result
    b.n1 <- sum(boo.A); b.n0 <- n- b.n1;
    boo.nume[b] <- sum(boo.data$Y*boo.A)/b.n1 - sum(boo.data$Y*(1-boo.A))/b.n0;
  }
  
  # Compute bootstrap test statistic
  TB <- abs(Y1.bar - Y0.bar - theta)/sd(boo.nume);
  rule<-(TB>qnorm(1-alpha/2));
  
   # List of returns
  list(stat=TB,rule=rule,cv=qnorm(1-alpha/2))

} 
#-------------------------------------------------------------------
shao.boot <- cmpfun(pre.shao.boot);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
true.values <- function(model,gamma,sg1,num.strata,pi){
#-------------------------------------------------------------------
# X-plain: true values of the DGP
#-------------------------------------------------------------------
  NN<-100000; # Sample size to compute varsigmas. 
  base.BIG  <- gen.baseline.one(NN,model,0,gamma,sg1);
  I.S       <-form.strata.one(base.BIG,num.strata);
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  Y.0 <-base.BIG$Y.0;
  Y.1 <-base.BIG$Y.1;

  # varsigma Y
  v.sg.Y <- var(Y.1)/pi + var(Y.0)/(1-pi); 
  v.sg.YB <- var(Y.1)/(1-pi) + var(Y.0)/pi;

  # varsigma tilde Y
  cond.mean.0 <- colMeans(Y.0*naI.S,na.rm=TRUE);
  cond.mean.1 <- colMeans(Y.1*naI.S,na.rm=TRUE);
  Vec.mean.0  <- rowSums(matrix(rep(cond.mean.0,each=NN),nrow=NN)*naI.S,na.rm=TRUE); 
  Vec.mean.1  <- rowSums(matrix(rep(cond.mean.1,each=NN),nrow=NN)*naI.S,na.rm=TRUE);
  tilde.Y.0   <- base.BIG$Y.0 - Vec.mean.0;
  tilde.Y.1   <- base.BIG$Y.1 - Vec.mean.1;
  v.sg.tilde.Y <- var(tilde.Y.1)/pi + var(tilde.Y.0)/(1-pi);
  v.sg.tilde.YB<- var(tilde.Y.1)/(1-pi) + var(tilde.Y.0)/pi;

  # varsigma H
  p.s <- colMeans(I.S);
  C.Mean.m0 <- colMeans(base.BIG$m.0*naI.S,na.rm=TRUE);
  C.Mean.m1 <- colMeans(base.BIG$m.1*naI.S,na.rm=TRUE);
  v.sg.H    <- sum(p.s*(C.Mean.m1-C.Mean.m0)^2);
  # varsigma A: this is zero for BCD and SBR
  v.sg.A.srs <- pi*(1-pi)*sum(p.s*(C.Mean.m1/pi+C.Mean.m0/(1-pi))^2);
  v.sg.A.srsB<- pi*(1-pi)*sum(p.s*(C.Mean.m1/(1-pi)+C.Mean.m0/pi)^2);
  v.sg.A.wei <- (1/3)*v.sg.A.srs;
  v.sg.A.weiB<- (1/3)*v.sg.A.srsB;
  # varsigma pi: This appears in sfe without saturation
  v.sg.pi.srs<- (1-2*pi)^2/(pi*(1-pi))*sum(p.s*(C.Mean.m1-C.Mean.m0)^2);
  # variance Numerator of t-statistics
  num        <- v.sg.tilde.Y+v.sg.H+v.sg.A.srs;
  
  list(VY1=var(Y.1),VY0=var(Y.0),VTY1=var(tilde.Y.1),VTY0=var(tilde.Y.0),
      vsgY=v.sg.Y,vsgTY=v.sg.tilde.Y,vsgH=v.sg.H,vsgA=v.sg.A.srs);
}

#-------------------------------------------------------------------
adjt.stat <- function(data,pi=1/2,tau=0,theta=0){
#-------------------------------------------------------------------
# X-plain: Computes the two sample t statistict with new SE
#-------------------------------------------------------------------
  Y   <- data$Y;
  A   <- data$A;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y.0 <- Y*(1-A);
  Y.1 <- Y*A;
  I.S <-data[,grepl("I.S", names(data))];
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  n.0.s<-colSums(naI.S*(1-A),na.rm=TRUE);
  n.1.s<-colSums(naI.S*A,na.rm=TRUE);
  sum.0.s<-colSums(Y.0*naI.S,na.rm=TRUE);
  sum.1.s<-colSums(Y.1*naI.S,na.rm=TRUE);
  mu.0.s <- sum.0.s/n.0.s;
  mu.1.s <- sum.1.s/n.1.s;

# Numerator
  Y1.bar   <-sum(Y.1)/n.1; 
  Y0.bar   <-sum(Y.0)/n.0;

# varsigma tilde Y
  p.s <- colMeans(I.S);
  term1<-(1/n.1)*sum(Y^2*A)-sum(p.s*mu.1.s^2,na.rm=TRUE);
  term0<-(1/n.0)*sum(Y^2*(1-A))-sum(p.s*mu.0.s^2,na.rm=TRUE);
  v.sg.tilde.Y <- term1/pi + term0/(1-pi);

# varsigma H  
  t1    <- sum(p.s*(mu.1.s-Y1.bar)^2,na.rm=TRUE);
  t2    <- sum(p.s*(mu.0.s-Y0.bar)^2,na.rm=TRUE);
  t3    <- 2*sum(p.s*(mu.1.s-Y1.bar)*(mu.0.s-Y0.bar),na.rm=TRUE);
  v.sg.H<- t1 + t2 - t3;

# The version below is worse with small samples and lots of NAs
#  v.sg.H    <- sum(p.s*(mu.1.s-mu.0.s+Y0.bar-Y1.bar)^2,na.rm=TRUE);

# varsigma A
  aux1  <- t1/(pi^2);
  aux2  <- t2/((1-pi)^2);
  aux3  <- t3/(pi*(1-pi));
  v.sg.A <- tau*(aux1+aux2+aux3); 

# The version below is worse with small samples and lots of NAs
  #vector<- (mu.1.s-Y1.bar)/pi + (mu.0.s-Y0.bar)/(1-pi); 
  #v.sg.A<- tau*sum(p.s*vector^2,na.rm=TRUE);

# denominator
denominator<- (v.sg.tilde.Y + v.sg.H + v.sg.A)/n;

adjt.stat<-abs(Y1.bar-Y0.bar-theta)/sqrt(denominator);

return(adjt.stat);
}

#-------------------------------------------------------------------
pre.CAP.t <- function(data,A.CAP,theta=0,alpha=0.05){
#-------------------------------------------------------------------
# X-plain: Covariate Adaptive Permutation test. 
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - A.CAP: n x M matrix of covariate adaptive permutations
#         - alpha: significance level of the test   
#         - theta: value of theta under the null
#-------------------------------------------------------------------
  I.S   <-data[,grepl("I.S", names(data))]
  num.strata<- ncol(I.S);
  Y   <-data$Y;
  n   <- length(Y);
  ObsT<-t.stat(data,theta);
  M   <-dim(A.CAP)[2];

  # Now Compute permuted test statistic. First get all transformations
  Y1.per <- Y*A.CAP; # every column is 1 permutation
  Y0.per <- Y*(1-A.CAP); # every column is 1 permutation
  n.1    <- colSums(A.CAP); n.0<-n-n.1; # number of ones and zeros per column
  bar_1  <- colSums(Y1.per)/n.1; # average per set of permutations
  bar_0  <- colSums(Y0.per)/n.0;
  vec.bar_1 <-t(matrix(rep(bar_1,n),M,n)); # useful to compute variance
  vec.bar_0 <-t(matrix(rep(bar_0,n),M,n));
  
  sig.sq_1   <- colSums((Y1.per-vec.bar_1)^2*A.CAP)/n.1;
  sig.sq_0   <- colSums((Y0.per-vec.bar_0)^2*(1-A.CAP))/n.0;
    
  # Compute the new test statistic over permuted data
  nume    <- abs(bar_1-bar_0-theta);
  deno    <- sqrt(sig.sq_1/n.1 + sig.sq_0/n.0);
  NewT    <- sort(nume/deno);    # sort the vector of Test Stat  
  k       <- M-floor(M*alpha);   # find index for quantile
  Mplus   <- sum(NewT>NewT[k]);  # M+ from LR book
  M0      <- sum(NewT==NewT[k]); # M0 from LR book
  prob    <- (M*alpha-Mplus)/M0; # term a(x) from LR book
  # randomized test 
  rule <- (ObsT>NewT[k]) + (ObsT==NewT[k])*(runif(1)<=prob);  
  # non-randomized test
  Nrule <- (ObsT>NewT[k]);
  ties <- sum(ObsT==NewT[k]);
  
  # List of returns
  list(stat=ObsT,rule=rule, Nrule=Nrule, cv=NewT[k],ties=ties)  
}
#-------------------------------------------------------------------
CAP.t <- cmpfun(pre.CAP.t);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
pre.CAP.t.adj <- function(data,PA,pi,tau,theta=0,alpha=0.05){
#-------------------------------------------------------------------
# X-plain: Covariate Adaptive Permutation test for the adjusted t-test 
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - PA: n x M matrix of covariate adaptive permutations
#         - pi: target proportion of treated
#         - tau: related to CAR mechanism
#         - alpha: significance level of the test   
#         - theta: value of theta under the null
#-------------------------------------------------------------------
  Y   <- data$Y;
  A   <- data$A;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y.0 <- Y*(1-A);
  Y.1 <- Y*A;
  I.S <-data[,grepl("I.S", names(data))];
  num.strata<- ncol(I.S);
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  n.0.s<-colSums(naI.S*(1-A),na.rm=TRUE);
  n.1.s<-colSums(naI.S*A,na.rm=TRUE);
  
  # Compute observed 
  ObsT<-adjt.stat(data,pi,tau);
  M   <-dim(PA)[2];
  p.s <- matrix(colMeans(I.S),num.strata,1);

  # Permuted versions: 
  Y1.per <- Y*PA; # every column is 1 permutation
  Y0.per <- Y*(1-PA); # every column is 1 permutation
  bar_1  <- colSums(Y1.per)/n.1;
  bar_0  <- colSums(Y0.per)/n.0;
  vbar_1  <- matrix(rep(colSums(Y1.per)/n.1,each=num.strata),num.strata,M); # average per set of permutations
  vbar_0  <- matrix(rep(colSums(Y0.per)/n.0,each=num.strata),num.strata,M);

  mu.1.s <- matrix(0,num.strata,M);
  mu.0.s <- matrix(0,num.strata,M);
  term.A1<-0;  term.A0<-0;
  term.B1<-0;  term.B0<-0; term.10 <-0;

  # For the new standard errors: we loop over number of strata
  for (s in 1:num.strata){
    mu.1.s[s,] <- colSums(Y1.per*naI.S[,s],na.rm=TRUE)/n.1.s[s];
    mu.0.s[s,] <- colSums(Y0.per*naI.S[,s],na.rm=TRUE)/n.0.s[s];
  } 

  term.A1 <- colSums(mu.1.s^2*as.vector(p.s),na.rm=TRUE);
  term.A0 <- colSums(mu.0.s^2*as.vector(p.s),na.rm=TRUE);
  term.B1 <- colSums((mu.1.s-vbar_1)^2*as.vector(p.s),na.rm=TRUE);
  term.B0 <- colSums((mu.0.s-vbar_0)^2*as.vector(p.s),na.rm=TRUE);
  term.10 <- colSums(2*(mu.1.s-vbar_1)*(mu.0.s-vbar_0)*as.vector(p.s),na.rm=TRUE);

  # varsigma tilde Y
  term1<-pmax((1/n.1)*colSums(Y^2*PA)-term.A1,0);
  term0<-pmax((1/n.0)*colSums(Y^2*(1-PA))-term.A0,0);
  v.sg.tilde.Y <- term1/pi + term0/(1-pi); # vector is 1 x M (one for each perm)

  # varsigma H  
  v.sg.H<- term.B0 + term.B1 - term.10;

  # varsigma A
  v.sg.A <- tau*(term.B1/(pi^2)+term.B0/((1-pi)^2)+term.10/(pi*(1-pi))); 
 
  # Compute the new test statistic over permuted data
  nume    <- abs(bar_1-bar_0-theta);
  deno    <- sqrt((v.sg.tilde.Y + v.sg.H + v.sg.A)/n);
  NewT    <- sort(nume/deno);    # sort the vector of Test Stat  
  k       <- M-floor(M*alpha);   # find index for quantile
  Mplus   <- sum(NewT>NewT[k]);  # M+ from LR book
  M0      <- sum(NewT==NewT[k]); # M0 from LR book
  prob    <- (M*alpha-Mplus)/M0; # term a(x) from LR book
  # randomized test 
  rule <- (ObsT>NewT[k]) + (ObsT==NewT[k])*(runif(1)<=prob);  
  # non-randomized test
  Nrule <- (ObsT>NewT[k]);
  ties <- sum(ObsT==NewT[k]);
  
  # List of returns
  list(stat=ObsT,rule=rule, Nrule=Nrule, cv=NewT[k],ties=ties)  
}
#-------------------------------------------------------------------
CAP.t.adj <- cmpfun(pre.CAP.t.adj);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
pre.CAP.sfe <- function(data,PA,pi,tau,theta=0,alpha=0.05){
#-------------------------------------------------------------------
# X-plain: Covariate Adaptive Permutation test for the adjusted t-test 
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - PA: n x M matrix of covariate adaptive permutations
#         - pi: target proportion of treated
#         - tau: related to CAR mechanism
#         - alpha: significance level of the test   
#         - theta: value of theta under the null
#-------------------------------------------------------------------
  Y   <- data$Y;
  A   <- data$A;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y.0 <- Y*(1-A);
  Y.1 <- Y*A;
  I.S <-data[,grepl("I.S", names(data))];
  num.strata<- ncol(I.S);
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  n.0.s<-colSums(naI.S*(1-A),na.rm=TRUE);
  n.1.s<-colSums(naI.S*A,na.rm=TRUE);
  M   <-dim(PA)[2];
  p.s <- matrix(colMeans(I.S),num.strata,1);
  ns  <- colSums(I.S);

  # Permuted versions: 
  Y1.per <- Y*PA; # every column is 1 permutation
  Y0.per <- Y*(1-PA); # every column is 1 permutation
  bar_1  <- matrix(rep(colSums(Y1.per)/n.1,each=num.strata),num.strata,M); # average per set of permutations
  bar_0  <- matrix(rep(colSums(Y0.per)/n.0,each=num.strata),num.strata,M);

  mu.1.s <- matrix(0,num.strata,M);
  mu.0.s <- matrix(0,num.strata,M);
  ratio  <- matrix(0,n,num.strata);
  term.A1<-0;  term.A0<-0;
  term.B1<-0;  term.B0<-0; term.10 <-0;

  # For the new standard errors: we loop over number of strata
  for (s in 1:num.strata){
    mu.1.s[s,] <- colSums(Y1.per*naI.S[,s],na.rm=TRUE)/n.1.s[s];
    mu.0.s[s,] <- colSums(Y0.per*naI.S[,s],na.rm=TRUE)/n.0.s[s];
    ratio[,s]  <- I.S[,s]*n.1.s[s]/ns[s];
  } 

  ratio <- matrix(rowSums(ratio,na.rm=TRUE),n,1);

  # Numerator
  tildeA <- PA-as.vector(ratio);
  top    <- colSums(tildeA*as.vector(Y),na.rm=TRUE);
  bottom <- colSums(tildeA^2,na.rm=TRUE);
  nume   <- abs(top/bottom-theta);

  term.A1 <- colSums(mu.1.s^2*as.vector(p.s),na.rm=TRUE);
  term.A0 <- colSums(mu.0.s^2*as.vector(p.s),na.rm=TRUE);
  term.B1 <- colSums((mu.1.s-bar_1)^2*as.vector(p.s),na.rm=TRUE);
  term.B0 <- colSums((mu.0.s-bar_0)^2*as.vector(p.s),na.rm=TRUE);
  term.10 <- colSums(2*(mu.1.s-bar_1)*(mu.0.s-bar_0)*as.vector(p.s),na.rm=TRUE);

  # varsigma tilde Y
  term1<-pmax((1/n.1)*colSums(Y^2*PA)-term.A1,0);
  term0<-pmax((1/n.0)*colSums(Y^2*(1-PA))-term.A0,0);
  v.sg.tilde.Y <- term1/pi + term0/(1-pi); # vector is 1 x M (one for each perm)

  # varsigma H  
  constant<-(1/(pi*(1-pi))-3);
  v.sg.H<- term.B0 + term.B1 - term.10;
  
  # Compute the new test statistic over permuted data
  deno    <- sqrt((v.sg.tilde.Y + constant*v.sg.H)/n);
  ObsT    <- nume[1]/deno[1];    # Recall first permutation is the identity
  NewT    <- sort(nume/deno);    # sort the vector of Test Stat  
  k       <- M-floor(M*alpha);   # find index for quantile
  Mplus   <- sum(NewT>NewT[k]);  # M+ from LR book
  M0      <- sum(NewT==NewT[k]); # M0 from LR book
  prob    <- (M*alpha-Mplus)/M0; # term a(x) from LR book
  # randomized test 
  rule <- (ObsT>NewT[k]) + (ObsT==NewT[k])*(runif(1)<=prob);  
  # non-randomized test
  Nrule <- (ObsT>NewT[k]);
  ties <- sum(ObsT==NewT[k]);
  
  # List of returns
  list(stat=ObsT,rule=rule, Nrule=Nrule, cv=NewT[k],ties=ties)  
}
#-------------------------------------------------------------------
CAP.sfe <- cmpfun(pre.CAP.sfe);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
pre.CAP.sfe.adj <- function(data,PA,pi,tau,theta=0,alpha=0.05){
#-------------------------------------------------------------------
# X-plain: Covariate Adaptive Permutation test for the adjusted t-test 
#         - data: entire data (covariates, treatment assigment, outcomes) 
#         - PA: n x M matrix of covariate adaptive permutations
#         - pi: target proportion of treated
#         - tau: related to CAR mechanism
#         - alpha: significance level of the test   
#         - theta: value of theta under the null
#-------------------------------------------------------------------
  Y   <- data$Y;
  A   <- data$A;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y.0 <- Y*(1-A);
  Y.1 <- Y*A;
  I.S <-data[,grepl("I.S", names(data))];
  num.strata<- ncol(I.S);
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  n.0.s<-colSums(naI.S*(1-A),na.rm=TRUE);
  n.1.s<-colSums(naI.S*A,na.rm=TRUE);
  M   <-dim(PA)[2];
  p.s <- matrix(colMeans(I.S),num.strata,1);
  ns  <- colSums(I.S);

  # Permuted versions: 
  Y1.per <- Y*PA; # every column is 1 permutation
  Y0.per <- Y*(1-PA); # every column is 1 permutation
  bar_1  <- matrix(rep(colSums(Y1.per)/n.1,each=num.strata),num.strata,M); # average per set of permutations
  bar_0  <- matrix(rep(colSums(Y0.per)/n.0,each=num.strata),num.strata,M);

  mu.1.s <- matrix(0,num.strata,M);
  mu.0.s <- matrix(0,num.strata,M);
  ratio  <- matrix(0,n,num.strata);
  term.A1<-0;  term.A0<-0;
  term.B1<-0;  term.B0<-0; term.10 <-0;

  # For the new standard errors: we loop over number of strata
  for (s in 1:num.strata){
    mu.1.s[s,] <- colSums(Y1.per*naI.S[,s],na.rm=TRUE)/n.1.s[s];
    mu.0.s[s,] <- colSums(Y0.per*naI.S[,s],na.rm=TRUE)/n.0.s[s];
    ratio[,s]  <- I.S[,s]*n.1.s[s]/ns[s];
  } 

  ratio <- matrix(rowSums(ratio,na.rm=TRUE),n,1);

  # Numerator
  tildeA <- PA-as.vector(ratio);
  top    <- colSums(tildeA*as.vector(Y),na.rm=TRUE);
  bottom <- colSums(tildeA^2,na.rm=TRUE);
  nume   <- abs(top/bottom-theta);

  term.A1 <- colSums(mu.1.s^2*as.vector(p.s),na.rm=TRUE);
  term.A0 <- colSums(mu.0.s^2*as.vector(p.s),na.rm=TRUE);
  term.B1 <- colSums((mu.1.s-bar_1)^2*as.vector(p.s),na.rm=TRUE);
  term.B0 <- colSums((mu.0.s-bar_0)^2*as.vector(p.s),na.rm=TRUE);
  term.10 <- colSums(2*(mu.1.s-bar_1)*(mu.0.s-bar_0)*as.vector(p.s),na.rm=TRUE);

  # varsigma tilde Y
  term1<-pmax((1/n.1)*colSums(Y^2*PA)-term.A1,0);
  term0<-pmax((1/n.0)*colSums(Y^2*(1-PA))-term.A0,0);
  v.sg.tilde.Y <- term1/pi + term0/(1-pi); # vector is 1 x M (one for each perm)

  # varsigma H  
  v.sg.H<- term.B0 + term.B1 - term.10;

  # varsigma pi
  constant<-(1-2*pi)^2/(pi^2*(1-pi)^2);

  v.sg.pi <- tau*constant*v.sg.H; 
 
  # Compute the new test statistic over permuted data
  deno    <- sqrt((v.sg.tilde.Y + v.sg.H + v.sg.pi)/n);
  ObsT    <- nume[1]/deno[1];    # Recall first permutation is the identity
  NewT    <- sort(nume/deno);    # sort the vector of Test Stat  
  k       <- M-floor(M*alpha);   # find index for quantile
  Mplus   <- sum(NewT>NewT[k]);  # M+ from LR book
  M0      <- sum(NewT==NewT[k]); # M0 from LR book
  prob    <- (M*alpha-Mplus)/M0; # term a(x) from LR book
  # randomized test 
  rule <- (ObsT>NewT[k]) + (ObsT==NewT[k])*(runif(1)<=prob);  
  # non-randomized test
  Nrule <- (ObsT>NewT[k]);
  ties <- sum(ObsT==NewT[k]);
  
  # List of returns
  list(stat=ObsT,rule=rule, Nrule=Nrule, cv=NewT[k],ties=ties,deno=deno,nume=nume)  
}
#-------------------------------------------------------------------
CAP.sfe.adj <- cmpfun(pre.CAP.sfe.adj);
#-------------------------------------------------------------------

#-------------------------------------------------------------------
SFE.var <- function(data,pi,tau){
#-------------------------------------------------------------------
# X-plain: Computes new standard errors for SFE
#        - data: entire data (covariates, treatment assigment, outcomes)
#        - pi: target proportion of treated (must be scalar)
#        - tau: related to CAR mechanism
#-------------------------------------------------------------------
  Y   <- data$Y;
  A   <- data$A;
  n   <- length(A);
  n.1 <- sum(A); 
  n.0 <- n-n.1;
  Y.0 <- Y*(1-A);
  Y.1 <- Y*A;
  I.S <-data[,grepl("I.S", names(data))];
  naI.S       <- I.S; # here we replace zeros with NA
  naI.S[naI.S==0]<-NA
  n.0.s<-colSums(naI.S*(1-A),na.rm=TRUE);
  n.1.s<-colSums(naI.S*A,na.rm=TRUE);
  sum.0.s<-colSums(Y.0*naI.S,na.rm=TRUE);
  sum.1.s<-colSums(Y.1*naI.S,na.rm=TRUE);
  mu.0.s <- sum.0.s/n.0.s;
  mu.1.s <- sum.1.s/n.1.s;
  Y1.bar <-sum(Y.1)/n.1; 
  Y0.bar <-sum(Y.0)/n.0;

# varsigma tilde Y
  p.s <- colMeans(I.S);
  term1<-max((1/n.1)*sum(Y^2*A)-sum(p.s*mu.1.s^2,na.rm=TRUE),0);
  term0<-max((1/n.0)*sum(Y^2*(1-A))-sum(p.s*mu.0.s^2,na.rm=TRUE),0);
  v.sg.tilde.Y <- term1/pi + term0/(1-pi);

# varsigma H  
  t1    <- sum(p.s*(mu.1.s-Y1.bar)^2,na.rm=TRUE);
  t2    <- sum(p.s*(mu.0.s-Y0.bar)^2,na.rm=TRUE);
  t3    <- 2*sum(p.s*(mu.1.s-Y1.bar)*(mu.0.s-Y0.bar),na.rm=TRUE);
  v.sg.H<- t1 + t2 - t3;

# varsigma pi
  constant<-(1-2*pi)^2/(pi^2*(1-pi)^2);
  v.sg.pi <- tau*constant*v.sg.H; 

  new.variance <- v.sg.tilde.Y + v.sg.H + v.sg.pi;

  return(new.variance);
}