#f Author: Ivan Canay | Northwestern University | April 2017
#-------------------------------------------------------------------------------------------------
# X-plain: Computes the simulations for inference under CAR
# Latest Revision: This file only computes the SFE test with new standard errors when pi=1/2
#-------------------------------------------------------------------------------------------------
# RETURNS:
rm(list=ls(all=TRUE))               # Clear all variables

source("Functions.Inference.CAR.R")
library(xtable)

#-------------------------------------------------------------------
pre.Inference.CAR <-function(myseed,n,model=1,gamma=1,sg1=1,pi=1/2,num.strata=10,theta=0,MC=5000){
#-------------------------------------------------------------------  
#IMPUTS: myseed: seed for the simulations 
#		 model: 1-4 see Functions.Inference.CAR.R
#        gamma: coefficient in potential outcomes equation
#        sg1: variance of error for potential outcome 1. 
#        pi: taget proportion in {0.5,0.3,0.7}
#        num.strata: will be 10 or 4
#        theta: value of theta = mu_1 - mu_0.
#        MC: Number of monte carlo simulations

# GENERATE DATA and SET PARAMETERS
# - - - - - - - - - - - - - - - - -
t.pi=round(pi, digits = 2);
dir.create("Tables")
nametext <-paste("Tables/SFE",n,"M",model,gamma,sg1,"pi",t.pi,myseed,"txt",sep=".");
dir.create("Latex")
nameLaTeX<-paste("Latex/SFE",n,"M",model,gamma,sg1,"pi",t.pi,myseed,"tex",sep=".");
dir.create("RData")
Rname 	 <-paste("RData/SFE",n,"M",model,gamma,sg1,"pi",t.pi,myseed,"Rdata",sep=".");
alpha   <-0.05;
cv<-qnorm(1-alpha/2);
set.seed(myseed);

# Set all rejection probabilities to zero
init<-matrix(0,MC,1);
rej.srs      <-init; rej.wei <-init; rej.bcd <-init; rej.sbr <-init;
rej.adjt.srs <-init; rej.adjt.wei <-init; rej.adjt.bcd <-init; rej.adjt.sbr <-init;
rej.shao.srs <-init; rej.shao.wei<-init; rej.shao.bcd<-init; rej.shao.sbr<-init; 
rej.per.srs  <-init; rej.per.wei <-init; rej.per.bcd <-init; rej.per.sbr <-init;
rej.reg.srs  <-init; rej.reg.wei <-init; rej.reg.bcd <-init; rej.reg.sbr <-init;
rej.SFE.srs  <-init; rej.SFE.wei<-init; rej.SFE.bcd<-init; rej.SFE.sbr<-init;
rej.SFEho.srs <-init; rej.SFEho.wei<-init; rej.SFEho.bcd<-init; rej.SFEho.sbr<-init;
rej.SFEhe.srs <-init; rej.SFEhe.wei<-init; rej.SFEhe.bcd<-init; rej.SFEhe.sbr<-init;
rej.CAP.srs   <-init; rej.CAP.wei <-init; rej.CAP.bcd <-init; rej.CAP.sbr <-init; 
rej.adjCAP.srs<-init; rej.adjCAP.wei <-init; rej.adjCAP.bcd <-init; rej.adjCAP.sbr <-init;
rej.SFEP.srs <-init; rej.SFEP.wei <-init; rej.SFEP.bcd <-init; rej.SFEP.sbr <-init;
rej.adjSFEP.srs <-init; rej.adjSFEP.wei <-init; rej.adjSFEP.bcd <-init; rej.adjSFEP.sbr <-init;

time1 <- proc.time()
time3 <- proc.time()

deep.param<-true.values(model,gamma,sg1,num.strata,pi);

for (j in 1:MC){
	# Generate Baseline covariates and potential outcomes (also Strata)
	base	<- gen.baseline.one(n,model,theta,gamma,sg1);
	I.S 	<- form.strata.one(base,num.strata);

	# Generate treatment assignments using CAR
	data.srs<-dgp.srs(base,I.S,pi);
	data.wei<-dgp.wei(base,I.S,pi);
	data.bcd<-dgp.bcd(base,I.S,pi);
	data.sbr<-dgp.sbr(base,I.S,pi);


	# (2) Adjusted Two Sample t-test
	wei.tau<- (1/(1+4*pi))*pi*(1-pi);
	

	# (4) Regression with Strata Fixed Effects (SFE)
  	SFE.srs <-lm(Y ~ A + I.S-1,data.srs);
	SFE.wei <-lm(Y ~ A + I.S-1,data.wei);
	SFE.bcd <-lm(Y ~ A + I.S-1,data.bcd);
	SFE.sbr <-lm(Y ~ A + I.S-1,data.sbr);
	nume.sfe.srs <-sqrt(n)*coef(summary(SFE.srs))[1,1];
	nume.sfe.wei <-sqrt(n)*coef(summary(SFE.wei))[1,1];
	nume.sfe.bcd <-sqrt(n)*coef(summary(SFE.bcd))[1,1];
	nume.sfe.sbr <-sqrt(n)*coef(summary(SFE.sbr))[1,1];


		wei.tau<- (1/(1+4*pi))*pi*(1-pi);
		srs.tau<- pi*(1-pi);
		var.sfe.srs  <-SFE.var(data.srs,pi,srs.tau);
		var.sfe.wei  <-SFE.var(data.wei,pi,wei.tau);
		var.sfe.bcd  <-SFE.var(data.bcd,pi,0);
		var.sfe.sbr  <-SFE.var(data.sbr,pi,0);

		rej.SFE.srs[j]<-(abs(nume.sfe.srs)/sqrt(var.sfe.srs)>cv);
		rej.SFE.wei[j]<-(abs(nume.sfe.wei)/sqrt(var.sfe.wei)>cv);
		rej.SFE.bcd[j]<-(abs(nume.sfe.bcd)/sqrt(var.sfe.bcd)>cv);
		rej.SFE.sbr[j]<-(abs(nume.sfe.sbr)/sqrt(var.sfe.sbr)>cv);

	# Report Iterations on Screen for monitoring
	if (floor(10*j/MC)==10*j/MC){
    cat("% completed = ", j/MC*100, "\n")
	  time4 <- proc.time();
    cat("Time Last round (min) = ", (time4[3]-time3[3])/60, "\n")
    time3 <-proc.time()
	} 
  
}

time2 <- proc.time();
cat(sprintf("Elapsed Time (min): %1.3f",(time2[3]-time1[3])/60), "\n");

## Display Information
basics_a <- c(model,n,theta,gamma,sg1,num.strata,alpha,t.pi,MC);
names(basics_a)<- c("model", "n  ", "theta", "gamma", "sg1", "#strata", "alpha", "pi", "MC iterations");

## Compute average rejections across simulations, ignoring possible NAS for methods that may return NAs
# original two-sample t-test
R.srs<-mean(rej.srs,na.rm=TRUE)*100;R.wei<-mean(rej.wei,na.rm=TRUE)*100;R.bcd<-mean(rej.bcd,na.rm=TRUE)*100;R.sbr<-mean(rej.sbr,na.rm=TRUE)*100;
# adjusted two sample t-test
R.adjt.srs<-mean(rej.adjt.srs,na.rm=TRUE)*100;R.adjt.wei<-mean(rej.adjt.wei,na.rm=TRUE)*100;R.adjt.bcd<-mean(rej.adjt.bcd,na.rm=TRUE)*100;R.adjt.sbr<-mean(rej.adjt.sbr,na.rm=TRUE)*100;
# adjusted CAP t-test
R.adjCAP.srs<-mean(rej.adjCAP.srs,na.rm=TRUE)*100;R.adjCAP.wei<-mean(rej.adjCAP.wei,na.rm=TRUE)*100;R.adjCAP.bcd<-mean(rej.adjCAP.bcd,na.rm=TRUE)*100;R.adjCAP.sbr<-mean(rej.adjCAP.sbr,na.rm=TRUE)*100; 
# CAP SFE test 
R.SFEP.srs<-mean(rej.SFEP.srs,na.rm=TRUE)*100;R.SFEP.wei<-mean(rej.SFEP.wei,na.rm=TRUE)*100;R.SFEP.bcd<-mean(rej.SFEP.bcd,na.rm=TRUE)*100;R.SFEP.sbr<-mean(rej.SFEP.sbr,na.rm=TRUE)*100;
# adjusted CAP SFE test 
R.adjSFEP.srs<-mean(rej.adjSFEP.srs,na.rm=TRUE)*100;R.adjSFEP.wei<-mean(rej.adjSFEP.wei,na.rm=TRUE)*100;R.adjSFEP.bcd<-mean(rej.adjSFEP.bcd,na.rm=TRUE)*100;R.adjSFEP.sbr<-mean(rej.adjSFEP.sbr,na.rm=TRUE)*100;


## Null rejection probabilities (for LATEX tables)
if (pi==1/2){
SRS.tex <-c(R.srs,R.adjt.srs,mean(rej.reg.srs)*100,mean(rej.SFE.srs)*100,mean(rej.CAP.srs)*100,R.adjCAP.srs,R.SFEP.srs);
WEI.tex <-c(R.wei,R.adjt.wei,mean(rej.reg.wei)*100,mean(rej.SFE.wei)*100,mean(rej.CAP.wei)*100,R.adjCAP.wei,R.SFEP.wei);
BCD.tex <-c(R.bcd,R.adjt.bcd,mean(rej.reg.bcd)*100,mean(rej.SFE.bcd)*100,mean(rej.CAP.bcd)*100,R.adjCAP.bcd,R.SFEP.bcd);
SBR.tex <-c(R.sbr,R.adjt.sbr,mean(rej.reg.sbr)*100,mean(rej.SFE.sbr)*100,mean(rej.CAP.sbr)*100,R.adjCAP.sbr,R.SFEP.sbr);
names(SRS.tex)  <-c("          t-test |", "     t-test*        |", "  Linear Reg. |", "SFE*   |", "   CAP     |", "   CAP* |", "   SFE-P |");
results <-rbind(SRS.tex,WEI.tex,BCD.tex,SBR.tex);
} else {
SRS.tex <-c(R.srs,R.adjt.srs,mean(rej.SFEhe.srs)*100,mean(rej.SFE.srs)*100,mean(rej.CAP.srs)*100,R.adjCAP.srs,R.SFEP.srs,R.adjSFEP.srs);
WEI.tex <-c(R.wei,R.adjt.wei,mean(rej.SFEhe.wei)*100,mean(rej.SFE.wei)*100,mean(rej.CAP.wei)*100,R.adjCAP.wei,R.SFEP.wei,R.adjSFEP.wei);
BCD.tex <-c(R.bcd,R.adjt.bcd,mean(rej.SFEhe.bcd)*100,mean(rej.SFE.bcd)*100,mean(rej.CAP.bcd)*100,R.adjCAP.bcd,R.SFEP.bcd,R.adjSFEP.bcd);
SBR.tex <-c(R.sbr,R.adjt.sbr,mean(rej.SFEhe.sbr)*100,mean(rej.SFE.sbr)*100,mean(rej.CAP.sbr)*100,R.adjCAP.sbr,R.SFEP.sbr,R.adjSFEP.sbr);
names(SRS.tex)  <-c("          t-test |", "     t-test*        |", "SFE   |", "SFE*   |", "   CAP     |", "   CAP* |", "   SFE-P |", "   SFE-P* |");
results <-rbind(SRS.tex,WEI.tex,BCD.tex,SBR.tex);
}
# Save LaTeX table
X.results<-cbind(c("SRS","WEI","BCD","SBR"),results)
row.names(X.results)<-c(model," ","  " ,"   " );
X.results<-xtable(X.results)
print(X.results, type="latex", file=nameLaTeX);


## Null rejection probabilities (for TXT Tables)
if (pi==1/2){
SRS <-c(R.srs,R.adjt.srs,mean(rej.reg.srs)*100,mean(rej.SFE.srs)*100,mean(rej.CAP.srs)*100,R.adjCAP.srs,R.SFEP.srs,mean(rej.per.srs)*100,mean(rej.SFEho.srs)*100,mean(rej.shao.srs)*100);
WEI <-c(R.wei,R.adjt.wei,mean(rej.reg.wei)*100,mean(rej.SFE.wei)*100,mean(rej.CAP.wei)*100,R.adjCAP.wei,R.SFEP.wei,mean(rej.per.wei)*100,mean(rej.SFEho.wei)*100,mean(rej.shao.wei)*100);
BCD <-c(R.bcd,R.adjt.bcd,mean(rej.reg.bcd)*100,mean(rej.SFE.bcd)*100,mean(rej.CAP.bcd)*100,R.adjCAP.bcd,R.SFEP.bcd,mean(rej.per.bcd)*100,mean(rej.SFEho.bcd)*100,mean(rej.shao.bcd)*100);
SBR <-c(R.sbr,R.adjt.sbr,mean(rej.reg.sbr)*100,mean(rej.SFE.sbr)*100,mean(rej.CAP.sbr)*100,R.adjCAP.sbr,R.SFEP.sbr,mean(rej.per.sbr)*100,mean(rej.SFEho.sbr)*100,mean(rej.shao.sbr)*100);
names(SRS)  <-c("            t-test |", "     t-test*      |", "  Linear Reg. |", "  SFE*      |", "     CAP      |", "     CAP*    |", "      SFE-P    |", "     Naive    |", "   SFE-homo    |", "     SYZ   |");
results <-rbind(SRS,WEI,BCD,SBR);
} else {
SRS <-c(R.srs,R.adjt.srs,mean(rej.SFEhe.srs)*100,mean(rej.SFE.srs)*100,mean(rej.CAP.srs)*100,R.adjCAP.srs,R.SFEP.srs,R.adjSFEP.srs,mean(rej.per.srs)*100,mean(rej.SFEho.srs)*100,mean(rej.shao.srs)*100,mean(rej.reg.srs)*100);
WEI <-c(R.wei,R.adjt.wei,mean(rej.SFEhe.wei)*100,mean(rej.SFE.wei)*100,mean(rej.CAP.wei)*100,R.adjCAP.wei,R.SFEP.wei,R.adjSFEP.wei,mean(rej.per.wei)*100,mean(rej.SFEho.wei)*100,mean(rej.shao.wei)*100,mean(rej.reg.wei)*100);
BCD <-c(R.bcd,R.adjt.bcd,mean(rej.SFEhe.bcd)*100,mean(rej.SFE.bcd)*100,mean(rej.CAP.bcd)*100,R.adjCAP.bcd,R.SFEP.bcd,R.adjSFEP.bcd,mean(rej.per.bcd)*100,mean(rej.SFEho.bcd)*100,mean(rej.shao.bcd)*100,mean(rej.reg.bcd)*100);
SBR <-c(R.sbr,R.adjt.sbr,mean(rej.SFEhe.sbr)*100,mean(rej.SFE.sbr)*100,mean(rej.CAP.sbr)*100,R.adjCAP.sbr,R.SFEP.sbr,R.adjSFEP.sbr,mean(rej.per.sbr)*100,mean(rej.SFEho.sbr)*100,mean(rej.shao.sbr)*100,mean(rej.reg.sbr)*100);
names(SRS)  <-c("             t-test | ", "    t-test*      |", "    SFE        |", "   SFE*   |", "     CAP      |", "      CAP*   |", "      SFE-P    |", "   SFE-P*     |", "    Naive      |", "   SFE-homo   |", "   SYZ      |", "   Lin-Reg  |");
results <-rbind(SRS,WEI,BCD,SBR);
}

print(results)

# # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Saving information to a text file
out <- file(nametext,"wt");   # Choose name of the file here
cat("-------------------------------------------------------------------------------------------------------------------------------------------------", "\n",file=out);
cat(date(), "\n",file=out)
cat("Seed used in this round = ", myseed , "\n",file=out)
cat("\n",file=out)
cat("----------------------------------------------- Parameters --------------------------------------------------------------------------------------", "\n",file=out);
cat(names(basics_a),sep="   ","\n",file=out);
cat(formatC(basics_a,digits=2,format="f"),sep="   ", "\n",file=out);
cat("\n",file=out)
  cat(c('Var of Y(1) =',round(deep.param$VY1,digits = 2)), "\n", file=out)
  cat(c('Var of Y(0) =',round(deep.param$VY0,digits = 2)), "\n", file=out)
  cat(c('Var of til Y(1) =',round(deep.param$VTY1,digits = 2)), "\n", file=out)
  cat(c('Var of til Y(0) =',round(deep.param$VTY0,digits = 2)), "\n", file=out)
  cat(c('Var of Y =',round(deep.param$vsgY,digits = 2) ), "\n", file=out)
  cat(c('Var of tilde Y =',round(deep.param$vsgTY,digits = 2) ), "\n", file=out)
  cat(c('Var of H =',round(deep.param$vsgH,digits = 2)), "\n", file=out)
  cat(c('Var of A =',round(deep.param$vsgA,digits = 2)), "\n", file=out)
cat("\n",file=out)
cat("-------------------------------------------------------------------------- REJECTION PROBABILITIES: SFE TEST --------------------------------------------------------------------", "\n",file=out);
cat(names(SRS),sep=" ","\n",file=out);
write.table(format(results, drop0trailing=FALSE),file=out,quote=F,sep='\t\t',row.names=T,col.names=F)
cat("\n",file=out)
cat("----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------", "\n",file=out);
time2 <- proc.time();
cat(sprintf("Elapsed Time (min): %1.3f",(time2[3]-time1[3])/60), "\n",file=out);
cat("------------------------------------------------------------------------------------------------------------------------------------------------", "\n",file=out);
cat(" SRS: simple random sampling. WEI: wei's design. BCD: bias coin design. SBR: Stratified Block Randomization ", "\n",file=out);
cat(" Model 1: linear model with normal epsi and beta W. ", "\n",file=out);
cat(" Model 2: (New) Heterogenous with logs  and beta W ", "\n",file=out);
cat(" Model 3: (previously 4) Uniform W and t-distributed error terms (old model 5 in paper) ", "ß\n",file=out);
cat(" Model 4: (previously 5) Same as model 5 but heterogenous (new model | model 4 in Expensive CAR paper) ", "\n",file=out);
cat("------------------------------------------------------------------------------------------------------------------------------------------------", "\n",file=out);
cat(" NOTE: SYZ was only computed for the BCD as it is not justified for other CAR mechanisms ", "\n",file=out);
cat("------------------------------------------------------------------------------------------------------------------------------------------------", "\n",file=out);
close(out);

save(list = ls(all = TRUE), file = Rname)

}
#-------------------------------------------------------------------
Inference.CAR <- cmpfun(pre.Inference.CAR);
#-------------------------------------------------------------------